Attribute VB_Name = "basMsgBoxEx"
'****************************************************************************************
'Module:    mMsgBoxEx - BAS Module
'Filename:  mMsgBoxEx.bas
'Author:    Jim Kahl
'Purpose:   to Hook the standard Windows MsgBox so we can alter it to display the MsgBox
'           the way we want it to look with custom icons, button captions, and font as
'           well as giving it a standard display position and time out period in the
'           event that the user does not respond within a certain period of time
'NOTES:
'   1 - This compiled DLL when referenced properly in a Project will become part of the
'       Global name space of the Project.  This means that the routines (of the Class
'       Module only) and the enumerations are fully accessible in the Project as if they
'       were part of Visual Basic
'   2 - Employs the use of a Resource File for possible Icons to be used
'   3 - HelpFile and Context parameters are for direct support of VB MsgBox function
'       ie. changing any existing MsgBox statement to a MsgBoxEx statement will have
'       no effect on the operation of the application
'   4 - Icon parameter uses Enumerated Constant jkMsgBoxIcons that are stored in the
'       resource file - this allows developer flexibility outside this module to use
'       predefined constants rather than try to remember the index numbers or path names
'       when trying to pass a valid Icon Handle to the routine
'   5 - Font parameter - this will change the font of the message box - since the
'       font used by the message box is a logical font the size increases are vastly
'       different from working screen fonts or printer fonts
'   6 - ButtonText parameter - to be passed in as a "Join"ed string using the "|"
'       character as a delimiter to change the caption of the buttons
'   7 - NoButtons parameter - allows us to show a message box that destroys the buttons
'       and disables the X button so the user can not dismiss the message box, the
'       message box is also resized appropriately (if this is specified as True then
'       the resulting message box is changed to a Timeout style message box)
'   8 - Timeout parameter - to close window after desired Interval in Seconds - if
'       you want a 3 1/4 second timeout period, code the parameter as 3.25
'   9 - DefRetVal parameter - to allow the developer to define the return value
'       in the event of a timeout, can be any of the vbMsgBoxResult values including
'       ones not defined in the Options to create the message box ie. vbYesNoCancel can
'       return vbOK in the event of a Timeout, this allows flexibility and gives the
'       developer a way to determine the difference between a Timeout event and a
'       user click event
'   10 - OpacityPercent - this is a percentage value that represents how opaque the
'       message box will appear to the user (Note: only 10 to 100% is available)
'       the smaller the number the more translucent the message box will appear
'****************************************************************************************
'****************************************************************************************
'****************************************************************************************
'*****                                  WARNING!!!!!!
'***** THIS MODULE EMPLOYS HOOKS AND SUBCLASSING - DO NOT CHANGE THIS MODULE UNLESS YOU
'***** KNOW WHAT YOU ARE DOING BECAUSE THE PROGRAM WILL CRASH UNEXPECTEDLY IF YOU DO
'***** SOMETHING INCORRECTLY
'****************************************************************************************
'****************************************************************************************
'****************************************************************************************
Option Explicit

'****************************************************************************************
'API CONSTANTS
'****************************************************************************************
'Layered Window Attributes
Private Const LWA_ALPHA As Long = &H2
'Window styles
Private Const GWL_EXSTYLE As Long = -20
Private Const WS_EX_LAYERED As Long = &H80000
'Menu flags
Private Const MF_BYPOSITION As Long = 1024
'Hooks
Private Const WH_CBT As Long = 5
Private Const HCBT_ACTIVATE As Long = 5
Private Const HWND_TOPMOST As Long = -1
'SetWindowPos
Private Const SWP_NOSIZE As Long = &H1
Private Const SWP_NOZORDER As Long = &H4
Private Const SWP_NOACTIVATE As Long = &H10
Private Const SWP_NOMOVE As Long = &H2
'WindowMessages
Private Const STM_SETICON As Long = &H170
Private Const WM_SETFONT As Long = &H30
'LogFont
Private Const LF_FACESIZE As Long = 32
'MsgBox Button ID's
Private Const IDOK As Long = 1
Private Const IDCANCEL As Long = 2
Private Const IDABORT As Long = 3
Private Const IDRETRY As Long = 4
Private Const IDIGNORE As Long = 5
Private Const IDYES As Long = 6
Private Const IDNO As Long = 7
Private Const IDHELP As Long = 9
'System Parameters
Private Const SPI_GETWORKAREA As Long = 48

'Constants commented out but left in place in case we want to experiment or use other
'values

'Font Clip Precision
'Private Const CLIP_CHARACTER_PRECIS As Long = 1
Private Const CLIP_DEFAULT_PRECIS As Long = 0
'Private Const CLIP_EMBEDDED As Long = 128
'Private Const CLIP_LH_ANGLES As Long = 16
'Private Const CLIP_MASK As Long = &HF
'Private Const CLIP_STROKE_PRECIS As Long = 2
'Private Const CLIP_TT_ALWAYS As Long = 32

'Font Family
'Private Const FF_DECORATIVE As Long = 80
Private Const FF_DONTCARE As Long = 0
'Private Const FF_MODERN As Long = 48
'Private Const FF_ROMAN As Long = 16
'Private Const FF_SCRIPT As Long = 64
'Private Const FF_SWISS As Long = 32

'Font Weight
'Private Const FW_BOLD As Long = 700
'Private Const FW_DONTCARE As Long = 0
'Private Const FW_EXTRABOLD As Long = 800
'Private Const FW_EXTRALIGHT As Long = 200
'Private Const FW_HEAVY As Long = 900
'Private Const FW_BLACK As Long = FW_HEAVY
'Private Const FW_LIGHT As Long = 300
'Private Const FW_MEDIUM As Long = 500
'Private Const FW_NORMAL As Long = 400
'Private Const FW_REGULAR As Long = FW_NORMAL
'Private Const FW_SEMIBOLD As Long = 600
'Private Const FW_DEMIBOLD As Long = FW_SEMIBOLD
'Private Const FW_THIN As Long = 100
'Private Const FW_ULTRABOLD As Long = FW_EXTRABOLD
'Private Const FW_ULTRALIGHT As Long = FW_EXTRALIGHT

'Font Output Precision
'Private Const OUT_CHARACTER_PRECIS As Long = 2
Private Const OUT_DEFAULT_PRECIS As Long = 0
'Private Const OUT_DEVICE_PRECIS As Long = 5
'Private Const OUT_OUTLINE_PRECIS As Long = 8
'Private Const OUT_PS_ONLY_PRECIS As Long = 10
'Private Const OUT_RASTER_PRECIS As Long = 6
'Private Const OUT_STRING_PRECIS As Long = 1
'Private Const OUT_STROKE_PRECIS As Long = 3
'Private Const OUT_TT_ONLY_PRECIS As Long = 7
'Private Const OUT_TT_PRECIS As Long = 4

Private Const LOGPIXELSY As Long = 90
Private Const MB_TOPMOST As Long = &H40000

'****************************************************************************************
'API TYPES
'****************************************************************************************
Private Type RECT
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
End Type

Private Type LOGFONT
    lfHeight As Long
    lfWidth As Long
    lfEscapement As Long
    lfOrientation As Long
    lfWeight As Long
    lfItalic As Byte
    lfUnderline As Byte
    lfStrikeOut As Byte
    lfCharSet As Byte
    lfOutPrecision As Byte
    lfClipPrecision As Byte
    lfQuality As Byte
    lfPitchAndFamily As Byte
    lfFaceName As String * LF_FACESIZE
End Type

Private Type Size
    cX As Long
    cY As Long
End Type

Private Type POINTAPI
    X As Long
    Y As Long
End Type

Private Type TEXTMETRIC
    tmHeight As Long
    tmAscent As Long
    tmDescent As Long
    tmInternalLeading As Long
    tmExternalLeading As Long
    tmAveCharWidth As Long
    tmMaxCharWidth As Long
    tmWeight As Long
    tmOverhang As Long
    tmDigitizedAspectX As Long
    tmDigitizedAspectY As Long
    tmFirstChar As Byte
    tmLastChar As Byte
    tmDefaultChar As Byte
    tmBreakChar As Byte
    tmItalic As Byte
    tmUnderlined As Byte
    tmStruckOut As Byte
    tmPitchAndFamily As Byte
    tmCharSet As Byte
End Type

'****************************************************************************************
'API FUNCTIONS
'****************************************************************************************
Private Declare Function CallNextHookEx Lib "user32" ( _
                ByVal hHook As Long, _
                ByVal nCode As Long, _
                ByVal wParam As Long, _
                ByVal lParam As Long) _
                As Long

Private Declare Function CreateFontIndirect Lib "gdi32.dll" _
        Alias "CreateFontIndirectA" ( _
                lpLogFont As LOGFONT) _
                As Long

Private Declare Function DeleteObject Lib "gdi32.dll" ( _
                ByVal hObject As Long) _
                As Long

Private Declare Function DestroyWindow Lib "user32.dll" ( _
                ByVal hwnd As Long) _
                As Long

Private Declare Function EndDialog Lib "user32.dll" ( _
                ByVal hDlg As Long, _
                ByVal nResult As Long) _
                As Long

Private Declare Function FindWindowEx Lib "user32" _
        Alias "FindWindowExA" ( _
                ByVal ParenthWnd As Long, _
                ByVal ChildhWnd As Long, _
                ByVal ClassName As String, _
                ByVal Caption As String) _
                As Long

Private Declare Function FreeLibrary Lib "kernel32.dll" (ByVal hLibModule As Long) As Long

Private Declare Function GetClassName Lib "user32" _
        Alias "GetClassNameA" ( _
                ByVal hwnd As Long, _
                ByVal lpClassName As String, _
                ByVal nMaxCount As Long) _
                As Long

Private Declare Function GetCurrentThreadId Lib "kernel32" () As Long

Private Declare Function GetDC Lib "user32.dll" ( _
                ByVal hwnd As Long) _
                As Long

Private Declare Function GetDesktopWindow Lib "user32" () As Long

Private Declare Function GetDeviceCaps Lib "gdi32.dll" ( _
                ByVal hdc As Long, _
                ByVal nIndex As Long) _
                As Long

Private Declare Function GetDlgItem Lib "user32.dll" ( _
                ByVal hDlg As Long, _
                ByVal nIDDlgItem As Long) _
                As Long

Private Declare Function GetParent Lib "user32" (ByVal hwnd As Long) As Long

Private Declare Function GetProcAddress Lib "kernel32.dll" ( _
                ByVal hModule As Long, _
                ByVal lpProcName As String) _
                As Long

Private Declare Function GetSystemMenu Lib "user32" ( _
                ByVal hwnd As Long, _
                ByVal bRevert As Long) _
                As Long

Private Declare Function GetTextExtentPoint32 Lib "gdi32.dll" _
        Alias "GetTextExtentPoint32A" ( _
                ByVal hdc As Long, _
                ByVal lpsz As String, _
                ByVal cbString As Long, _
                lpSize As Size) _
                As Long

Private Declare Function GetTextMetrics Lib "gdi32.dll" _
        Alias "GetTextMetricsA" ( _
                ByVal hdc As Long, _
                lpMetrics As TEXTMETRIC) As Long

Private Declare Function GetWindowLong Lib "user32.dll" _
        Alias "GetWindowLongA" ( _
                ByVal hwnd As Long, _
                ByVal nIndex As Long) _
                As Long

Private Declare Function GetWindowRect Lib "user32" ( _
                ByVal hwnd As Long, _
                lpRect As RECT) _
                As Long

Private Declare Function KillTimer Lib "user32" ( _
                ByVal hwnd As Long, _
                ByVal nIDEvent As Long) _
                As Long

Private Declare Function LoadLibrary Lib "kernel32.dll" _
        Alias "LoadLibraryA" ( _
                ByVal lpLibFileName As String) _
                As Long

Private Declare Function MoveWindow Lib "user32.dll" ( _
                ByVal hwnd As Long, _
                ByVal X As Long, _
                ByVal Y As Long, _
                ByVal nWidth As Long, _
                ByVal nHeight As Long, _
                ByVal bRepaint As Long) _
                As Long

Private Declare Function ReleaseDC Lib "user32.dll" ( _
                ByVal hwnd As Long, _
                ByVal hdc As Long) _
                As Long

Private Declare Function RemoveMenu Lib "user32" ( _
                ByVal hMenu As Long, _
                ByVal nPosition As Long, _
                ByVal wFlags As Long) _
                As Long

Private Declare Function ScreenToClient Lib "user32" ( _
                ByVal hwnd As Long, _
                lpPoint As POINTAPI) As Long

Private Declare Function SelectObject Lib "gdi32.dll" ( _
                ByVal hdc As Long, _
                ByVal hObject As Long) _
                As Long

Private Declare Function SendMessage Lib "user32" _
        Alias "SendMessageA" ( _
                ByVal hwnd As Long, _
                ByVal wMsg As Long, _
                ByVal wParam As Long, _
                lParam As Any) _
                As Long

Private Declare Function SetDlgItemText Lib "user32" _
        Alias "SetDlgItemTextA" ( _
                ByVal hDlg As Long, _
                ByVal nIDDlgItem As Long, _
                ByVal lpString As String) _
                As Long

Private Declare Function SetLayeredWindowAttributes Lib "user32.dll" ( _
                ByVal hwnd As Long, _
                ByVal crKey As Long, _
                ByVal bAlpha As Byte, _
                ByVal dwFlags As Long) _
                As Long

Private Declare Function SetTimer Lib "user32" ( _
                ByVal hwnd As Long, _
                ByVal nIDEvent As Long, _
                ByVal uElapse As Long, _
                ByVal lpTimerFunc As Long) _
                As Long

Private Declare Function SetWindowLong Lib "user32.dll" _
        Alias "SetWindowLongA" ( _
                ByVal hwnd As Long, _
                ByVal nIndex As Long, _
                ByVal dwNewLong As Long) _
                As Long

Private Declare Function SetWindowPos Lib "user32" ( _
                ByVal hwnd As Long, _
                ByVal hWndInsertAfter As Long, _
                ByVal X As Long, _
                ByVal Y As Long, _
                ByVal cX As Long, _
                ByVal cY As Long, _
                ByVal wFlags As Long) _
                As Long

Private Declare Function SetWindowsHookEx Lib "user32" _
        Alias "SetWindowsHookExA" ( _
                ByVal idHook As Long, _
                ByVal lpfn As Long, _
                ByVal hMod As Long, _
                ByVal dwThreadId As Long) _
                As Long

Private Declare Function SetWindowText Lib "user32.dll" _
        Alias "SetWindowTextA" ( _
                ByVal hwnd As Long, _
                ByVal lpString As String) _
                As Long

Private Declare Function SystemParametersInfo Lib "user32.dll" _
        Alias "SystemParametersInfoA" ( _
                ByVal uAction As Long, _
                ByVal uParam As Long, _
                ByVal lpvParam As Any, _
                ByVal fuWinIni As Long) _
                As Long

Private Declare Function UnhookWindowsHookEx Lib "user32" ( _
                ByVal hHook As Long) _
                As Long

'****************************************************************************************
'CONSTANTS - PRIVATE
'****************************************************************************************
Private Const JDK_DEFAULT As Long = &HFFFFFFFF '-1 is reserved for centering
Private Const JDK_DEFAULT_CAPTION As String = "APP_TITLE"
Private Const ID_TMR = &H9000

'****************************************************************************************
'VARIABLES - PRIVATE
'****************************************************************************************
Private mStdFont As StdFont     'new standard font
Private mhLF As Long            'handle to the logical font
Private mhWnd As Long           'handle to the message box window
Private mlTmrID As Long         'timer id
Private msPrompt As String
Private mlHook As Long          'hook id
Private mlLeft As Long
Private mlTop As Long
Private mhIcon As Long          'icon handle
Private mlOptions As Long
Private msButtonText() As String
Private mbNoButtons As Boolean
Private mfTimeout As Single
Private mDefRetVal As VbMsgBoxResult
Private mlOpacity As Long
Private meRetVal As VbMsgBoxResult

'****************************************************************************************
'METHODS - PUBLIC
'****************************************************************************************
Public Function MsgBoxExtended(ByRef Prompt As String, _
                Optional ByVal Options As VbMsgBoxStyle, _
                Optional ByRef Title As String, _
                Optional ByRef HelpFile As String, _
                Optional ByVal Context As Long, _
                Optional ByVal Left As Long = -1, _
                Optional ByVal Top As Long = -1, _
                Optional ByVal Icon As Long, _
                Optional ByVal Font As StdFont, _
                Optional ByRef ButtonText As String = "||||", _
                Optional ByVal NoButtons As Boolean = False, _
                Optional ByVal Timeout As Single = -1, _
                Optional ByVal DefRetVal As VbMsgBoxResult = -1, _
                Optional ByVal OpacityPercent As Long = 100) As VbMsgBoxResult
    'Effects:       Starts the subclassing of the message box and sets up a timer
    '               if necessary
    'Parameters:    Prompt - the message to appear in the MsgBox
    '               Options - the standard MsgBox options - if using a custom icon
    '                   do not set the icon in these options
    '               Title - the caption in the title bar of the MsgBox
    '               HelpFile - filespec of the help file to use when the Help button
    '                   is displayed
    '               Context - the help context ID of the Help topic to be displyed
    '               Left - the left position of the MsgBox in respect to the owner form
    '               Top - the left position of the MsgBox in respect to the owner form
    '               Icon - a valid handle to an Icon resource
    '               Font - changes the font of the message area only
    '               ButtonText - the text to be displayed on the buttons in order
    '                   in some cases the second and third elements of the string
    '                   are ignored, the fourth element always applys to the Help button
    '               NoButtons - Allows us to create a message box that the user can't
    '                   dismiss - intended for use as a Timeout message box, if there
    '                   is no timeout specified, the Timeout will be set for 5 seconds
    '               Timeout - the time in seconds to dismiss the MsgBox - still has
    '                   millisecond precision ie. can be passed as 5.125
    '               DefRetVal - specifies the value to return when a Timeout event
    '                   occurs this value does not have to be one of the normal
    '                   return values can be any developer defined value
    'Returns:       standard MsgBox return values - ie. if using vbYesNoCancel and the
    '                   user clicks the second button the return value is vbNo - this
    '                   is regardless of what the caption states
    'NOTES:         The first 5 Parameters are the same as the standard MsgBox this is
    '                   so that the MsgBoxEx function can operate as a direct replacement
    '                   for current MsgBox statements - changing all MsgBox statements to
    '                   MsgBoxEx statements will not affect operation
    
    Dim hInst As Long
    Dim lThreadID As Long
    Dim picIco As New StdPicture
    Dim lRet As Long
    
    meRetVal = 0
    
    If ButtonText <> vbNullString Then
        'retrieve possible custom button captions
        msButtonText = Split(ButtonText, "|")
    End If
    
    hInst = App.hInstance
    lThreadID = GetCurrentThreadId()
    
    'Save the new arguments as member variables to be used from the MsgBoxHook proc
    mlLeft = Left
    mlTop = Top
    mlOptions = Options
    msPrompt = Prompt
    mDefRetVal = DefRetVal
    mbNoButtons = NoButtons
    mfTimeout = Timeout
    'make sure we have an opacity value between 10% and 100%
    Select Case OpacityPercent
        Case Is < 10
            mlOpacity = 10
        Case Is > 100
            mlOpacity = 100
        Case Else
            mlOpacity = OpacityPercent
    End Select
    
    'if a font was passed store it for later
    If Font Is Nothing Then
    Else
        Set mStdFont = Font
    End If
    
    'default the MsgBox caption to app.title
    If Title = JDK_DEFAULT_CAPTION Then
        Title = App.Title
    End If
    
    'if we don't want to display the buttons then change to just "OK" button
    If mbNoButtons Then
'        mlOptions = mlOptions And &HFFBFF0
        'need to make sure that a Timeout has been set in case of no buttons
        If mfTimeout <= 0 Then
            mfTimeout = 5
        End If
    End If
    
    'make sure that we are setting mhIcon to its proper value
    'either handle to an icon or 0
    If Icon <> 0& Then
        'this MUST be a valid icon - cannot be a bitmap
        Set picIco = LoadResPicture(Icon, vbResIcon)
        mhIcon = picIco.Handle
        'remove the current icon option
        mlOptions = mlOptions And &HFFFF8F
        'set icon to a known value so we can replace it
        mlOptions = mlOptions Or vbCritical
    Else
        mhIcon = 0&
    End If
    
    '"subclass" the MsgBox function
    mlHook = SetWindowsHookEx(WH_CBT, AddressOf msgBoxHook, hInst, lThreadID)
    
    'show the Message Box and let the hook process take care of the rest
    lRet = MsgBox(msPrompt, mlOptions + MB_TOPMOST, Title, HelpFile, Context)
    If meRetVal = 0 Then
        meRetVal = lRet
    End If
    MsgBoxExtended = meRetVal
    
    'after the message box is dismissed we need to delete our logfont
    Call DeleteObject(mhLF)
End Function

'****************************************************************************************
'METHODS - PRIVATE
'****************************************************************************************
Private Function msgBoxHook( _
        ByVal nCode As Long, _
        ByVal wParam As Long, _
        ByVal lParam As Long) As Long
    'Effects:   performs the hook of the MsgBox function to allow our custom changes to
    '           take effect
    'Inputs:    nCode - for our purposes here the only one that we care about will be
    '               the HCBT_ACTIVATE code
    '           wParam - for our purposes here we are only going to be looking at this as
    '               a handle to the message box
    '           lParam - for this function we are basically going to ignore this param
    'Returns:   Long integer representing success or failure of the hook
    Dim lHgt As Long
    Dim lWid As Long
    Dim lSize As Long
    Dim lHgtMb As Long
    Dim lWidMb As Long
    Dim rcMsgBox As RECT
    Dim rcOwnFrm As RECT
    Dim rcMsg As RECT
    Dim rcScrn As RECT
    Dim sBuff As String
    Dim lPosX As Long
    Dim lPosY As Long
    Dim cX As Long
    Dim cY As Long
    Dim X As Long
    Dim Y As Long
    Dim hwnd As Long
    Dim lf As LOGFONT
    Dim hdc As Long
    Dim sz As Size
    Dim sCR() As String
    Dim sWord() As String
    Dim lCnt As Long
    Dim pt As POINTAPI
    Dim tm As TEXTMETRIC
    Dim sLine As String
    Dim lCnt1 As Long
    Dim lLines As Long
    Const WIDTHPCT As Single = 0.8
    
    'NOTE: Class names we are interested in are as follows
    '#32770 - cbt dialog
    'Static - Icon
    'Button - Command Button

    'Call next hook in the chain and return the value
    '(this is the polite way to allow other hooks to function too)
    msgBoxHook = CallNextHookEx(mlHook, nCode, wParam, lParam)
        
    'hook only the activate msg
    If nCode = HCBT_ACTIVATE Then
        'handle only standard MsgBox class windows
        'this is the most efficient method to allocate strings in VB
        'according to Brad Martinez's results with tools from NuMega
        sBuff = Space$(32)

        'GetClassName will truncate the class name if it doesn't fit in the buffer
        'we only care about the first 6 chars anyway
        lSize = GetClassName(wParam, sBuff, 32)
        If Left$(sBuff, lSize) <> "#32770" Then
            'not a standard msgBox
            'we can just quit because we already called CallNextHookEx
            Exit Function
        End If

        'get the handle for the MsgBox
'        mhWnd = FindWindowEx(0&, 0&, "#32770", vbNullString)
        mhWnd = wParam
        
        'now that we have the handle to the MB we can set the timeout period
        If mfTimeout > 0 Then
            mfTimeout = mfTimeout * 1000
            mlTmrID = SetTimer(mhWnd, ID_TMR, mfTimeout, AddressOf timerProc)
        End If

        'store MsgBox size
        Call GetWindowRect(mhWnd, rcMsgBox)
        lHgtMb = rcMsgBox.Bottom - rcMsgBox.Top
        lWidMb = rcMsgBox.Right - rcMsgBox.Left
        lHgt = lHgtMb / 2
        lWid = lWidMb / 2
        
        'store parent window size
        Call GetWindowRect(GetParent(mhWnd), rcOwnFrm)
        
        'center msgbox to parent form
        lPosY = rcOwnFrm.Top + (rcOwnFrm.Bottom - rcOwnFrm.Top) / 2
        lPosX = rcOwnFrm.Left + (rcOwnFrm.Right - rcOwnFrm.Left) / 2

        'if user passed in specific values then use those instead
        If mlLeft = JDK_DEFAULT Then
            X = lPosX - lWid
        Else
            X = mlLeft
        End If

        If mlTop = JDK_DEFAULT Then
            Y = lPosY - lHgt
        Else
            Y = mlTop
        End If

        'If user passed in custom icon use that instead of the standard Windows icon
        If mhIcon <> 0& Then
            hwnd = FindWindowEx(mhWnd, 0&, "Static", vbNullString)
            Call SendMessage(hwnd, STM_SETICON, mhIcon, ByVal 0&)
        End If

        If mStdFont Is Nothing Then
            'properly position when not changing fonts
            SetWindowPos mhWnd, HWND_TOPMOST, X, Y, 0&, 0&, _
                    SWP_NOZORDER + SWP_NOACTIVATE + SWP_NOSIZE
        Else
            'a font name was passed in set the font to the new one and resize
            
            'get handle and find rectangle for message
            hwnd = FindWindowEx(mhWnd, 0&, "Static", msPrompt)
            Call GetWindowRect(hwnd, rcMsg)
                        
            'calulate the size difference between the message box and the message area
            'for use in calculating the size of the new message box
            cX = ((rcMsgBox.Right - rcMsg.Right) + (rcMsg.Left - rcMsgBox.Left))
            cY = ((rcMsgBox.Bottom - rcMsg.Bottom) + (rcMsg.Top - rcMsgBox.Top))
            
            'create the font we want to use
            lf = logFontFromStdFont(mStdFont)
            mhLF = CreateFontIndirect(lf)
            
            'set the font and output the text
            Call SendMessage(hwnd, WM_SETFONT, mhLF, True)
            
            'get the DC of the message window so we can manipulate it
            hdc = GetDC(hwnd)
            
            'select the new font into the dc so we can obtain our new text metrics
            Call SelectObject(hdc, mhLF)
            Call SetWindowText(hwnd, msPrompt)
            
            'append a carriage return to the end of the string before making our
            'split to ensure that all lines are accounted for
            msPrompt = msPrompt & vbCrLf
            sCR = Split(msPrompt, vbCrLf)
            
            'retrieve the coordinates of the screen
            SystemParametersInfo SPI_GETWORKAREA, 0, VarPtr(rcScrn), 0
            
            'reset lWid to be able to determine the length of the longest line
            lWid = 0
            sz.cX = 0
            
            Do
                'now find the amount of room the new font will need
                For lCnt = LBound(sCR) To UBound(sCR)
                    lLines = UBound(sCR) '+ 1
                    Call GetTextExtentPoint32(hdc, sCR(lCnt), Len(sCR(lCnt)), sz)
                    If lWid < sz.cX Then
                        lWid = sz.cX
                        lHgt = sz.cY
                    End If
                
                    If lWid > (rcScrn.Right - rcScrn.Left) * WIDTHPCT Then
                        'if the longest line is larger than 2/3 size of display area
                        'we want to break it up and rearrange it so the message box
                        'doesn't get too big
                        sLine = sCR(lCnt)
                        sWord = Split(sLine, " ")
                        sLine = vbNullString
                        For lCnt1 = LBound(sWord) To UBound(sWord)
                            sLine = sLine & sWord(lCnt1) & Space$(1)
                            GetTextExtentPoint32 hdc, sLine, Len(sLine), sz
'                            Debug.Print sz.cX, sz.cY
                            If sz.cX > (rcScrn.Right - rcScrn.Left) * WIDTHPCT Then
                                sLine = sLine & vbCrLf
                                lWid = (rcScrn.Right - rcScrn.Left) * WIDTHPCT
                                lLines = lLines + 1
                                Exit For
                            End If
                        Next lCnt1
                    End If
                Next lCnt
            Loop Until lCnt > UBound(sCR)
            
            'since lHgt is the height for one line of text, multiply by the total
            'number of lines to get the overall height of the message area
            lHgt = lHgt * lLines
            
            'lWid is the width needed to ensure the longest line doesn't get clipped,
            'well almost
            Debug.Print lWid; lHgt

            'now retrieve text metrics to find the overhang - this really only applies
            'to Italic fonts, without this some fonts might get clipped at the end of
            'the line
            'also need to get the descent so that the height includes parts of characters
            'that fall below the line
            Call GetTextMetrics(hdc, tm)
            Call ReleaseDC(hwnd, hdc)
            Call SetWindowPos(hwnd, 0&, 0&, 0&, lWid + tm.tmOverhang, _
                    lHgt + tm.tmDescent, SWP_NOMOVE + SWP_NOACTIVATE)
            
            'get the size of the message area and add our offsets to calculate the new
            'size of our message box
            Call GetWindowRect(hwnd, rcMsg)
            lWid = rcMsg.Right + cX
            lHgt = rcMsg.Bottom + cY

            'set dimensions to PointAPI structure and convert back to client units
            pt.X = lWid
            pt.Y = lHgt
            Call ScreenToClient(mhWnd, pt)
            
            'make sure the message box does not become too small or too large
            If pt.X < lWidMb Then
                pt.X = lWidMb
            End If
            If pt.Y < lHgtMb Then
                pt.Y = lHgtMb
            End If
                        
            'set the message box position and size
            SetWindowPos mhWnd, HWND_TOPMOST, X, Y, pt.X, pt.Y, _
                    SWP_NOZORDER + SWP_NOACTIVATE

        End If
        
        'this will resize the message box appropriately when there are no
        'buttons being displayed
        If mbNoButtons Then
            destroyButtons
            Call GetWindowRect(hwnd, rcMsg)
            Call GetWindowRect(mhWnd, rcMsgBox)
            SetWindowPos mhWnd, HWND_TOPMOST, 0&, 0&, rcMsgBox.Right - rcMsgBox.Left, _
                    rcMsg.Bottom - rcMsgBox.Top, _
                    SWP_NOZORDER + SWP_NOACTIVATE + SWP_NOMOVE
            GoTo ErrHandler
        Else
            setButtonPos
        End If
        
        'change the captions of the buttons
        setButtonCaptions

        'set the opacity of the message box
        Call SetWindowLong(mhWnd, GWL_EXSTYLE, _
                GetWindowLong(mhWnd, GWL_EXSTYLE) + WS_EX_LAYERED)
        If apiExists("user32", "SetLayeredWindowAttributes") Then
            Call SetLayeredWindowAttributes(mhWnd, 0, getTransValue, LWA_ALPHA)
        Else
        End If

ErrHandler:
        'unhook the dialog and we are out clean!
        UnhookWindowsHookEx mlHook
    End If
End Function

Private Sub setButtonCaptions()
    'Effects:   changes the caption on the buttons, some we can do directly by checking
    '           ID and setting the new caption others we have to see what style the
    '           message box is before we attempt to change the caption
    
    'no need to set captions for no buttons
    If mbNoButtons Then Exit Sub
    
    'check the various buttons and change their caption to our passed string parts
    
    'Special note: when attempting to get the handle of buttons with captions
    'Yes, No, Abort, Retry or Ignore with FindWindowEx what we are actually
    'looking for is &Yes, &No, etc. the captions are shown that way in Win95 but
    'not in XP, not sure about other OS Versions
        
    'OK, Yes, and Abort are always in the first position so we don't need to check what
    'style they are in just change them
    If msButtonText(0) <> vbNullString Then
        Call SetDlgItemText(mhWnd, IDOK, msButtonText(0))
        Call SetDlgItemText(mhWnd, IDYES, msButtonText(0))
        Call SetDlgItemText(mhWnd, IDABORT, msButtonText(0))
    End If
    
    If UBound(msButtonText) >= 1 Then
        'No is always in the second position
        If msButtonText(1) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDNO, msButtonText(1))
        End If
    Else
        Exit Sub
    End If
    
    If UBound(msButtonText) >= 2 Then
        'Ignore is always in the third position
        If msButtonText(2) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDIGNORE, msButtonText(2))
        End If
    Else
        Exit Sub
    End If
    
    If UBound(msButtonText) >= 3 Then
        'Help is considered the 4th button regardless of how many there actually are
        If msButtonText(3) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDHELP, msButtonText(3))
        End If
    Else
        Exit Sub
    End If
    
    'Now we handle the buttons that don't show up in the same place all the time
    
    'Retry/Cancel - note that the third element of the string is ignored
    If (mlOptions And vbRetryCancel) = vbRetryCancel Then
        If msButtonText(0) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDRETRY, msButtonText(0))
        End If
        If msButtonText(1) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDCANCEL, msButtonText(1))
        End If
    'Yes/No/Cancel
    ElseIf (mlOptions And vbYesNoCancel) = vbYesNoCancel Then
        'we already handled the Yes and No buttons
        If msButtonText(2) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDCANCEL, msButtonText(2))
        End If
    'Abort/Retry/Ignore
    ElseIf (mlOptions And vbAbortRetryIgnore) = vbAbortRetryIgnore Then
        'we already handled the Abort and Ignore buttons
        If msButtonText(1) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDRETRY, msButtonText(1))
        End If
    'OK/Cancel - note that the third element of the string is ignored in this case
    ElseIf (mlOptions And vbOKCancel) = vbOKCancel Then
        'we already handled the OK button
        If msButtonText(1) <> vbNullString Then
            Call SetDlgItemText(mhWnd, IDCANCEL, msButtonText(1))
        End If
    End If
    
End Sub

Private Function timerProc( _
        ByVal hwnd As Long, _
        ByVal uMsg As Long, _
        ByVal idEvent As Long, _
        ByVal dwTime As Long) As Long
    'Effects:   Closes the dialog and returns a default value at the end of the
    '           timeout period
    'Inputs:    hwnd - handle to the window that owns the timer
    '           uMsg - the WM_TIMER message
    '           idEvent - unique identifier of the timer
    '           dwTime - same value as GetTickCount (in milliseconds)
    'Returns:   these are specific to the message box button options chosen if the
    '           DefRetVal parameter is not set to a specifed value
    '           vbOKOnly - vbOK
    '           vbYesNo - vbNo
    '           vbAbortRetryIgnore - vbIgnore
    '           vbOKCancel - vbCancel
    '           vbYesNoCancel - vbCancel
    '           vbRetryCancel - vbCancel
    'Kill timer.
    If mlTmrID Then
        Call KillTimer(mhWnd, ID_TMR)
        mlTmrID = 0
    End If
    
    'set the return value and end the message
    If mDefRetVal <> JDK_DEFAULT Then
        'return the value defined by the developer
        meRetVal = mDefRetVal
    Else
        If (mlOptions And vbRetryCancel) = vbRetryCancel Then
            'for Retry/Cancel MsgBox's return a Cancel
            meRetVal = vbCancel
        ElseIf (mlOptions And vbYesNo) = vbYesNo Then
            'For Yes/No MsgBox's return a No
            meRetVal = vbNo
        ElseIf (mlOptions And vbYesNoCancel) = vbYesNoCancel Then
            'For Yes/No/Cancel MsgBox's return a Cancel
            meRetVal = vbCancel
        ElseIf (mlOptions And vbAbortRetryIgnore) = vbAbortRetryIgnore Then
            'For Abort/Retry/Ignore MsgBox's return an Ignore
            meRetVal = vbIgnore
        Else
            'For anything else just return a Cancel
            meRetVal = vbCancel
        End If
    End If
    
    EndDialog mhWnd, meRetVal
End Function

Private Sub setButtonPos()
    Dim lBtnCnt As Long
    Dim hwnd(0 To 3) As Long    'handle to one of the dialog buttons
    Dim hWndRef As Long         'handle to the reference button
    Dim rcRef As RECT           'rect of the reference button
    Dim rc As RECT              'rect of a dialog button
    Dim rcMB As RECT            'rect of the message box dialog
    Dim bHelp As Boolean
    Dim lOff As Long
    Dim lWid As Long
    Dim lHgt As Long
    Dim lLeft As Long
    Dim lTop As Long
    Dim lMidX As Long
    Dim lMidY As Long
    
    'get the size of the msgbox
    Call GetWindowRect(mhWnd, rcMB)
    
    'get the size of the message area
    lMidX = (rcMB.Right - rcMB.Left) / 2
    lMidY = (rcMB.Bottom - rcMB.Top)
    
    'now we want to determine the type of message box to find the number of buttons as
    'well as determine the button to use for our positioning reference
    If (mlOptions And vbRetryCancel) = vbRetryCancel Then
        lBtnCnt = 2
        hwnd(0) = GetDlgItem(mhWnd, IDRETRY)
        hwnd(1) = GetDlgItem(mhWnd, IDCANCEL)
        hWndRef = hwnd(1)
    ElseIf (mlOptions And vbYesNo) = vbYesNo Then
        lBtnCnt = 2
        hwnd(0) = GetDlgItem(mhWnd, IDYES)
        hwnd(1) = GetDlgItem(mhWnd, IDNO)
        hWndRef = hwnd(1)
    ElseIf (mlOptions And vbYesNoCancel) = vbYesNoCancel Then
        lBtnCnt = 3
        hwnd(0) = GetDlgItem(mhWnd, IDYES)
        hwnd(1) = GetDlgItem(mhWnd, IDNO)
        hwnd(2) = GetDlgItem(mhWnd, IDCANCEL)
        hWndRef = hwnd(1)
    ElseIf (mlOptions And vbAbortRetryIgnore) = vbAbortRetryIgnore Then
        lBtnCnt = 3
        hwnd(0) = GetDlgItem(mhWnd, IDABORT)
        hwnd(1) = GetDlgItem(mhWnd, IDRETRY)
        hwnd(2) = GetDlgItem(mhWnd, IDIGNORE)
        hWndRef = hwnd(1)
    ElseIf (mlOptions And vbOKCancel) = vbOKCancel Then
        lBtnCnt = 2
        hwnd(0) = GetDlgItem(mhWnd, IDOK)
        hwnd(1) = GetDlgItem(mhWnd, IDCANCEL)
        hWndRef = hwnd(1)
    Else
        lBtnCnt = 1
        hwnd(0) = GetDlgItem(mhWnd, IDOK)
        hWndRef = hwnd(0)
    End If
    
    hwnd(3) = GetDlgItem(mhWnd, IDHELP)
    If hwnd(3) <> 0 Then
        bHelp = True
    End If
    
    'get the rectangle of our reference button
    Call GetWindowRect(hWndRef, rcRef)
    
    'set variables that will not change (offset, width, height, etc.)
    If lBtnCnt <> 1 Then
        Call GetWindowRect(hwnd(0), rc)
        lOff = rcRef.Left - rc.Left
    Else
        lOff = 5
    End If
    lWid = rcRef.Right - rcRef.Left
    lHgt = rcRef.Bottom - rcRef.Top
    lLeft = lMidX - (lWid / 2)
    lTop = lMidY - (lHgt * 3)
    
    Select Case lBtnCnt
        Case 1  'OK Only
            If bHelp Then
                Call MoveWindow(hwnd(0), lLeft - lOff - lWid / 2, lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(3), lLeft + lOff + lWid / 2, lTop, lWid, lHgt, True)
            Else
                Call MoveWindow(hwnd(0), lLeft, lTop, lWid, lHgt, True)
            End If
        Case 2  'Ok/Cancel, Retry/Cancel or Yes/No
            If bHelp Then
                Call MoveWindow(hwnd(0), lLeft - lOff, lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(1), lLeft, lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(3), lLeft + lOff, lTop, lWid, lHgt, True)
            Else
                Call MoveWindow(hwnd(0), lLeft - (lOff - lWid / 2), lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(1), lLeft + lWid / 2, lTop, lWid, lHgt, True)
            End If
        Case 3  'Yes/No/Cancel or Abort/Retry/Ignore
            If bHelp Then
                Call MoveWindow(hwnd(0), lLeft - lOff - (lOff - lWid / 2), lTop, lWid, _
                        lHgt, True)
                Call MoveWindow(hwnd(1), lLeft - (lOff - lWid / 2), lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(2), lLeft + (lWid / 2), lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(3), lLeft + (lWid / 2) + lOff, lTop, lWid, lHgt, True)
            Else
                Call MoveWindow(hwnd(0), lLeft - lOff, lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(1), lLeft, lTop, lWid, lHgt, True)
                Call MoveWindow(hwnd(2), lLeft + lOff, lTop, lWid, lHgt, True)
            End If
    End Select
End Sub

Private Function logFontFromStdFont(ByRef sf As StdFont) As LOGFONT
    'Effects:   message box uses a logical font so we need to retrieve the std font
    '           passed to the MsgBoxEx routine and convert it
    'Inputs:    sf - an StdFont that we want to use for the message box
    'Returns:   a logical font that can be used to place our message into the hDC of
    '           the message box
    Dim fnt As LOGFONT
    
    'creates a LOGFONT from an StdFont
    With fnt
        .lfCharSet = sf.Charset
        .lfItalic = sf.Italic
        .lfFaceName = sf.NAME & vbNullChar
        .lfHeight = pointsToHeight(sf.Size)
        .lfStrikeOut = sf.Strikethrough
        .lfUnderline = sf.Underline
        .lfWeight = sf.Weight
        .lfClipPrecision = CLIP_DEFAULT_PRECIS
        .lfOutPrecision = OUT_DEFAULT_PRECIS
        'note this does not work for some of the decorative fonts
        .lfPitchAndFamily = FF_DONTCARE
    End With

    logFontFromStdFont = fnt
End Function

Private Function pointsToHeight(ByVal Points As Long) As Single
    'Effectss:  Converts a StdFont point size to an equivalent LOGFONT lfHeight value
    'Inputs:    Points - the size in points of the StdFont to use
    'Returns:   a number that will equate the point size to the proper height
    pointsToHeight = -(Points * pixelsPerInchY) / 72
End Function

Private Function pixelsPerInchY() As Long
    'Returns:   number of pixels per logical inch of the screen
    Dim hwnd As Long
    Dim hdc As Long
    'Retrieves the number of pixels per logical inch on the screens Y axis
    hwnd = GetDesktopWindow()
    hdc = GetDC(hwnd)
    pixelsPerInchY = GetDeviceCaps(hdc, LOGPIXELSY)
    Call ReleaseDC(hwnd, hdc)
    
End Function

Private Sub destroyButtons()
    'Purpose:   destroys the buttons so that the user can not dismiss the message box
    Dim hMenu As Long
    Dim lCnt As Long
    Dim hwnd As Long
    
    'when no buttons are displayed, disable the X button and remove the
    'close menu item on the system menu so the user cannot dismiss it that way
    hMenu = GetSystemMenu(mhWnd, 0)
    Call RemoveMenu(hMenu, 1, MF_BYPOSITION)
    
    'ignore errors for now
    On Error Resume Next
    
    'loop through the possible buttons and destroy them when we find them
    For lCnt = IDOK To IDHELP
        hwnd = GetDlgItem(mhWnd, lCnt)
        Call DestroyWindow(hwnd)
    Next lCnt
    
    'resume normal error handling
    On Error GoTo 0
End Sub

Private Function getTransValue() As Long
    'Purpose:   converts the passed OpacityPercent value from a percentage to a long so
    '           it can be used directly with the SetLayeredWindowAttributes call
    Dim lValue As Long
    
    Const STEP As Single = 2.55
    lValue = mlOpacity * STEP
    getTransValue = lValue
End Function

Private Function apiExists(ByRef Module As String, ByRef FuncName As String) As Boolean
    'Purpose:   determines if an API function is available on the user pc.
    'Inputs:    Module - the name of the exe or dll that would contain the sought after
    '               function
    '           FuncName - the name of the API function
    'Returns:   True - the specified API function exists in the library
    '           False - either the Module was not found or the function was not found
    '               in the module
    Dim hMod As Long
    Dim lpAddr As Long
    
    hMod = LoadLibrary(Module)
    If hMod Then
        lpAddr = GetProcAddress(hMod, FuncName)
        apiExists = CBool(lpAddr)
        FreeLibrary hMod
    End If
End Function








